@@ -4,8 +4,10 @@
 Control flow graph builder.
 """
 
-from typing import List, Optional
+from collections import defaultdict, deque
+from typing import Dict, List, Optional, DefaultDict, Deque, Tuple, Union
 import ast  # type: ignore
+import enum
 from _ast import (
     Import,
     Break,
@@ -25,11 +27,15 @@ from _ast import (
     stmt,
     While,
     Yield,
+    Try,
+    ExceptHandler,
 )  # type: ignore
-from py2cfg.model import Block, Link, CFG, FuncBlock
+from py2cfg.model import Block, TryBlock, Link, CFG, FuncBlock
 
 
-def invert(node: Compare) -> Compare:
+def invert(
+    node: Union[Compare, ast.expr]
+) -> Union[ast.NameConstant, ast.UnaryOp, Compare]:
     """
     Invert the operation in an ast node object (get its negation).
 
@@ -54,7 +60,7 @@ def invert(node: Compare) -> Compare:
 
     if isinstance(node, ast.Compare):
         op = type(node.ops[0])
-        inverse_node = ast.Compare(
+        inverse_node: Union[ast.NameConstant, ast.UnaryOp, Compare] = ast.Compare(
             left=node.left, ops=[inverse[op]()], comparators=node.comparators
         )
     elif isinstance(node, ast.NameConstant) and node.value in [True, False]:
@@ -65,7 +71,10 @@ def invert(node: Compare) -> Compare:
     return inverse_node
 
 
-def merge_exitcases(exit1: Optional[Compare], exit2: None) -> Optional[Compare]:
+def merge_exitcases(
+    exit1: Union[Compare, ast.BoolOp, ast.expr, None],
+    exit2: Union[Compare, ast.BoolOp, ast.expr, None],
+) -> Union[Compare, ast.BoolOp, ast.expr, None]:
     """
     Merge the exitcases of two Links.
 
@@ -83,6 +92,31 @@ def merge_exitcases(exit1: Optional[Compare], exit2: None) -> Optional[Compare]:
     return exit2
 
 
+class TryEnum(enum.IntEnum):
+    BODY = enum.auto()
+    EXCEPT = enum.auto()
+    ELSE = enum.auto()
+    FINAL = enum.auto()
+
+
+class TryStackObject:
+    def __init__(
+        self,
+        try_block: TryBlock,
+        after_block: Block,
+        has_final: bool,
+        iter_state: Optional[TryEnum] = None,
+    ) -> None:
+        self.try_block = try_block
+        self.after_block = after_block
+        self.has_final = has_final
+        self.iter_state = iter_state
+
+    @property
+    def node(self) -> ast.Try:
+        return self.try_block.statements[0]  # type: ignore
+
+
 class CFGBuilder(ast.NodeVisitor):
     """
     Control flow graph builder.
@@ -91,9 +125,20 @@ class CFGBuilder(ast.NodeVisitor):
     a program's AST and iteratively build the corresponding CFG.
     """
 
-    def __init__(self, short: bool = True) -> None:
+    def __init__(
+        self, short: bool = True, treebuf: DefaultDict[str, Deque] = None
+    ) -> None:
         self.isShort = short
-        self._callbuf: FuncBlock = []
+        self._callbuf: List[FuncBlock] = []
+        self._treebuf = defaultdict(deque) if treebuf is None else treebuf
+
+    @property
+    def loop_stack(self):
+        return self._treebuf["loop_stack"]
+
+    @property
+    def try_stack(self) -> Deque[TryStackObject]:
+        return self._treebuf["try_stack"]
 
     # ---------- CFG building methods ---------- #
     def build(
@@ -155,7 +200,7 @@ class CFGBuilder(ast.NodeVisitor):
             return self.build_from_src(name, src)
 
     # ---------- Graph management methods ---------- #
-    def new_block(self) -> Block:
+    def new_block(self, statement=None) -> Block:
         """
         Create a new block with a new id.
 
@@ -163,7 +208,10 @@ class CFGBuilder(ast.NodeVisitor):
             A Block object with a new unique id.
         """
         self.current_id += 1
-        return Block(self.current_id)
+        block = Block(self.current_id)
+        if statement is not None:
+            block.add_statement(statement)
+        return block
 
     def new_func_block(self) -> FuncBlock:
         """
@@ -175,7 +223,14 @@ class CFGBuilder(ast.NodeVisitor):
         self.current_id += 1
         return FuncBlock(self.current_id)
 
-    def add_statement(self, block: Block, statement: stmt) -> None:
+    def new_try_block(self, statement=None):
+        self.current_id += 1
+        block = TryBlock(self.current_id)
+        if statement is not None:
+            block.add_statement(statement)
+        return block
+
+    def add_statement(self, block: Block, statement: Union[stmt, Call]) -> None:
         """
         Add a statement to a block.
 
@@ -187,7 +242,10 @@ class CFGBuilder(ast.NodeVisitor):
         block.statements.append(statement)
 
     def add_exit(
-        self, block: Block, nextblock: Block, exitcase: Optional[Compare] = None
+        self,
+        block: Block,
+        nextblock: Block,
+        exitcase: Union[Compare, None, ast.BoolOp, ast.expr] = None,
     ) -> None:
         """
         Add a new exit to a block.
@@ -236,7 +294,7 @@ class CFGBuilder(ast.NodeVisitor):
         # A new sub-CFG is created for the body of the class definition and
         # added to the class CFGs of the current CFG.
         class_body = ast.Module(body=node.body)
-        class_builder = CFGBuilder(short=self.isShort)
+        class_builder = CFGBuilder(self.isShort, self._treebuf)
         self.cfg.classcfgs[node.name] = class_builder.build(
             node.name, class_body, asynchr, self.current_id
         )
@@ -256,11 +314,13 @@ class CFGBuilder(ast.NodeVisitor):
         # A new sub-CFG is created for the body of the function definition and
         # added to the function CFGs of the current CFG.
         func_body = ast.Module(body=node.body)
-        func_builder = CFGBuilder(short=self.isShort)
-        self.cfg.functioncfgs[node.name] = func_builder.build(
+        func_builder = CFGBuilder(self.isShort, self._treebuf)
+        cfg = self.cfg.functioncfgs[node.name] = func_builder.build(
             node.name, func_body, asynchr, self.current_id
         )
         self.current_id = func_builder.current_id + 1
+        cfg.lineno = node.lineno
+        cfg.end_lineno = node.end_lineno
 
     def clean_cfg(self, block: Block, visited: List[Block] = []) -> None:
         """
@@ -322,11 +382,21 @@ class CFGBuilder(ast.NodeVisitor):
                 func_name += "." + node.attr
                 return func_name
             elif isinstance(node, ast.Subscript):
-                return node.value.id
+                if "id" in node.value._fields:
+                    return node.value.id
+                elif "attr" in node.value._fields:
+                    return visit_func(node.value)
+                else:
+                    raise AttributeError("WTF is this thing, build it in??", type(node))
             elif isinstance(node, ast.Constant):
                 return node.value
             elif isinstance(node, ast.Call):
-                return node.func.id
+                if "id" in node.func._fields:
+                    return node.func.id
+                elif "attr" in node.func._fields:
+                    return visit_func(node.func)
+                else:
+                    raise AttributeError("WTF is this thing, build it in??", type(node))
             else:
                 print("WTF is this thing, build it in??", type(node))
 
@@ -366,8 +436,87 @@ class CFGBuilder(ast.NodeVisitor):
         self.generic_visit(node)
 
     def visit_Raise(self, node):
-        # TODO
-        pass
+        if self.current_block.statements:
+            raise_block = self.new_block(node)
+            self.current_block.add_exit(raise_block)
+            self.current_block = raise_block
+        else:
+            self.current_block.add_statement(node)
+
+        if not self.try_stack:
+            # Raise statement outside of try block
+            # If we don't know where control jumps, this is the last block
+            self.current_block = self.new_block(node)
+            return
+
+        if isinstance(node.exc, ast.Call):
+            e_id = node.exc.func.id
+        elif isinstance(node.exc, ast.Name):
+            e_id = node.exc.id
+        else:
+            raise ValueError(f"Unexpected object {node.exc}")
+
+        for tryobj in list(self.try_stack):
+
+            def contains(item, state):
+                return (
+                    item in tryobj.try_block.except_blocks
+                    and tryobj.iter_state == state
+                )
+
+            if contains(e_id, TryEnum.BODY):
+                # try:
+                #     raise StopIteration
+                # except StopIteration:
+                #     control_transfer_here = True
+                # except Exception:
+                #     control_transfer_here = False
+                self.current_block.add_exit(tryobj.try_block.except_blocks[e_id])
+                break
+            elif contains(None, TryEnum.BODY):
+                # try:
+                #     raise StopIteration
+                # except:
+                #     control_transfers_here = 1
+                self.current_block.add_exit(tryobj.try_block.except_blocks[None])
+                break
+
+            elif contains(e_id, TryEnum.EXCEPT) or contains(None, TryEnum.EXCEPT):
+                if tryobj.has_final:
+                    _after_block = self.new_block()
+                    self.current_block.add_exit(_after_block)
+                    _after_block.add_exit(self.current_block)
+                    self.current_block = _after_block
+
+                    for child in tryobj.node.finalbody:
+                        if isinstance(child, ast.Raise):
+                            break
+                        self.visit(child)
+
+            elif tryobj.iter_state == TryEnum.ELSE:
+                if tryobj.has_final:
+                    _after_block = self.new_block()
+                    self.current_block.add_exit(_after_block)
+                    _after_block.add_exit(self.current_block)
+                    self.current_block = _after_block
+                    for child in tryobj.node.finalbody:
+                        if isinstance(child, ast.Raise):
+                            break
+                        self.visit(child)
+
+            elif tryobj.iter_state != TryEnum.FINAL and tryobj.has_final:
+                # try: raise Exception
+                # finally: pass
+                _after_block = self.new_block()
+                self.current_block.add_exit(_after_block)
+                self.current_block = _after_block
+
+                for child in tryobj.node.finalbody:
+                    if isinstance(child, ast.Raise):
+                        top = self.try_stack.popleft()
+                        break
+                    self.visit(child)
+        self.current_block = self.new_block()
 
     def visit_Assert(self, node):
         self.add_statement(self.current_block, node)
@@ -405,7 +554,7 @@ class CFGBuilder(ast.NodeVisitor):
         afterif_block = self.new_block()
 
         # New block for the body of the else if there is an else clause.
-        if len(node.orelse) != 0:
+        if node.orelse:
             else_block = self.new_block()
             self.add_exit(self.current_block, else_block, invert(node.test))
             self.current_block = else_block
@@ -426,6 +575,8 @@ class CFGBuilder(ast.NodeVisitor):
         self.current_block = afterif_block
 
     def visit_While(self, node: While) -> None:
+        # TODO while/else
+
         loop_guard = self.new_loopguard()
         self.current_block = loop_guard
         self.add_statement(self.current_block, node)
@@ -442,14 +593,19 @@ class CFGBuilder(ast.NodeVisitor):
 
         # Populate the while block.
         self.current_block = while_block
+        self.loop_stack.appendleft((afterwhile_block, loop_guard))
         for child in node.body:
             self.visit(child)
+        top = self.loop_stack.popleft()
+        assert top == (afterwhile_block, loop_guard)
         self.add_exit(self.current_block, loop_guard)
 
         # Continue building the CFG in the after-while block.
         self.current_block = afterwhile_block
 
     def visit_For(self, node: For) -> None:
+        # TODO for/else
+
         loop_guard = self.new_loopguard()
         self.current_block = loop_guard
         self.add_statement(self.current_block, node)
@@ -465,21 +621,37 @@ class CFGBuilder(ast.NodeVisitor):
         self.add_exit(self.current_block, afterfor_block)
         self.current_block = for_block
 
+        # Push exit destinations for break/continue statements.
+        # On break, control jumps to afterfor_block.
+        # On continue, control jumps to loop_guard.
+        self.loop_stack.appendleft((afterfor_block, loop_guard))
+
         # Populate the body of the for loop.
         for child in node.body:
             self.visit(child)
+
+        top = self.loop_stack.popleft()
+        assert top == (afterfor_block, loop_guard)
         self.add_exit(self.current_block, loop_guard)
 
         # Continue building the CFG in the after-for block.
         self.current_block = afterfor_block
 
     def visit_Break(self, node: Break) -> None:
-        # TODO
-        pass
+        assert self.loop_stack
+        after_block: Block
+        after_block, _ = self.loop_stack[0]
+        self.current_block.add_statement(node)
+        self.current_block.add_exit(after_block)
+        self.current_block = self.new_block()
 
     def visit_Continue(self, node):
-        # TODO
-        pass
+        assert self.loop_stack
+        loop_guard: Block
+        _, loop_guard = self.loop_stack[0]
+        self.add_statement(self.current_block, node)
+        self.add_exit(self.current_block, loop_guard)
+        self.current_block = self.new_block()
 
     def visit_Import(self, node: Import) -> None:
         self.add_statement(self.current_block, node)
@@ -502,7 +674,23 @@ class CFGBuilder(ast.NodeVisitor):
         self.current_block = afterawait_block
 
     def visit_Return(self, node: Return) -> None:
+        if self.current_block.statements:
+            return_block = self.new_block()
+            self.current_block.add_exit(return_block)
+            self.current_block = return_block
+
         self.add_statement(self.current_block, node)
+
+        if self.try_stack:
+            stackobj = self.try_stack[0]
+            if stackobj.iter_state != TryEnum.FINAL and stackobj.has_final:
+                after_return = self.new_block()
+                self.current_block.add_exit(after_return)
+                after_return.add_exit(self.current_block)
+                self.current_block = after_return
+                for child in stackobj.node.finalbody:
+                    self.visit(child)
+
         self.cfg.finalblocks.append(self.current_block)
         # Continue in a new block but without any jump to it -> all code after
         # the return statement will not be included in the CFG.
@@ -513,3 +701,57 @@ class CFGBuilder(ast.NodeVisitor):
         afteryield_block = self.new_block()
         self.add_exit(self.current_block, afteryield_block)
         self.current_block = afteryield_block
+
+    def visit_Try(self, node: Try):
+        try_block = self.new_try_block(node)
+        after_tryblock = self.new_block()
+        self.current_block.add_exit(try_block)
+        stackobj = TryStackObject(try_block, after_tryblock, bool(node.finalbody))
+        self.try_stack.appendleft(stackobj)
+
+        stackobj.iter_state = TryEnum.EXCEPT
+        for child in node.handlers:
+            self.current_block = self.new_block()
+            # If we encounter a raise statement during body iteration,
+            # we can link the raise block to the appropriate exception block (if any).
+            try_block.except_blocks[
+                None if child.type is None else child.type.id
+            ] = self.current_block
+            self.visit(child)
+
+        stackobj.iter_state = TryEnum.BODY
+        self.current_block = try_block
+        for child in node.body:
+            self.visit(child)
+
+        if node.orelse:
+            stackobj.iter_state = TryEnum.ELSE
+            else_block = self.new_block()
+            self.current_block.add_exit(else_block)
+            self.current_block = else_block
+            for child in node.orelse:
+                self.visit(child)
+
+        self.current_block.add_exit(after_tryblock)
+        self.current_block = after_tryblock
+        if node.finalbody:
+            stackobj.iter_state = TryEnum.FINAL
+            for child in node.finalbody:
+                if isinstance(child, ast.Raise):
+                    top = self.try_stack.popleft()
+                    self.visit_Raise(child)
+                    self.try_stack.appendleft(top)
+                    break
+                self.visit(child)
+            else:
+                next_block = self.new_block()
+                self.current_block.add_exit(next_block)
+                self.current_block = next_block
+
+        del self.try_stack[0]
+
+    def visit_ExceptHandler(self, node: ExceptHandler):
+        assert self.try_stack
+        for child in node.body:
+            self.visit(child)
+        self.current_block.add_exit(self.try_stack[0].after_block)
